/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-06 * new file - Rolf
*/


//a class to interpret a memory region as a sequence of variable size bit strings
//the maximum supported symbol length is 25 bits
class
BitStream
{
private:
	BitStream();
public:
	BitStream(ubyte* data,uint size);
	~BitStream() {};

	uint peek()const {	return(  (*((uint*)data))>>(bit_offset)); }

	void advance(ubyte count) {		bit_offset+=count;
  		while (bit_offset>7) { ++data; bit_offset-=8; } }

//	uint extract(uint count) {	uint ret=peek(); advance(count); return(ret); }

	bool over()const {return(data>=max_byte); }

private:
	ubyte* data;
	ubyte* max_byte;
	ubyte bit_offset;
};

//class to accumulate variable length bitstrings
//the maximum supported symbol length is 25 bits
//clients of the class must maintain a storage pointer
//this pointer will be updated by the class when reallocation occurs
class
BitSink
{
private:
	BitSink();
public:
	BitSink(ubyte* &storage_pointer);
	~BitSink() {};

	void insert(uint symbol,uint sym_size);		//insert symbol of size 'sym_size' bits
	uint finalize();	//reduce allocated memory block to its final size and return number of bytes

private:
	ubyte* data;
	ubyte** client_data;
	uint byte_offset;
	uint bit_offset;
	uint allocated_bytes;
};

