/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

//stuff for understanding headers
#include <vector>           //needed for MaskedImage.h
#include <deque>			//needed for Generator.h and Boxed.h
#include "common.h"
#include "ExtremeFloat.h"	//needed for Fractal.h
#include "Section.h"		//needed for MaskedImage.h
#include "Span.h"			//needed for MaskedImage.h
#include "MaskedImage.h"	//needed for Fractal.h
#include "Fractal.h"		//needed for Generator.h
#include "Generator.h"   	//needed for Boxed.h
#include "Boxed.h"

//internal stuff
#include "x86_timer.ah"		//needed for Timing.h"
#include "Timing.h"

BoxedGenerator::BoxedGenerator(Fractal* fractal):
Generator(fractal)
{
	reset();
}

void
BoxedGenerator::reset()
{
	width=fractal->get_width();
	height=fractal->get_height();

	//empty sections stack
	sections.clear();
	//push rectangle of window dimensions onto section stack
	sections.push_back(Section(0,0,width,height));
	reset_preview();
	filled_samples=0;
}

bool
BoxedGenerator::refine_image(float max_time)
{
	fractal->ready();
	const MaskedImage* img=fractal->get_image();
	ushort x,y;

	AbsTimer refine_timer;

	bool ret=preview(max_time);

	while ((!sections.empty())&&(refine_timer.elapsed_seconds()<max_time))
	{
		//remove one rectangular section from queue
		Section s=sections.front();
		sections.pop_front();
		ret=true;

		uint delta_y=s.y1-s.y0;		//get (height-1) of rectangle
		uint delta_x=s.x1-s.x0;		//get (width-1) of rectangle

		//don't bother with lines or single points
		//they're included in the edges of 'real' rectangles anyway
		if ((delta_y*delta_x)==0) continue;

		if (delta_y==1)
		{
			//rectangle is exactly two samples high
			//splitting won't help, so we calculate all samples
			for (x=s.x0;x<=s.x1;++x)
			{
				fractal->sample(x,s.y0);
				fractal->sample(x,s.y1);
			}
			continue;
		}
		if (delta_x==1)
		{
			//rectangle is exactly two samples wide
			for (y=s.y0;y<=s.y1;++y)
			{
				fractal->sample(s.x0,y);
				fractal->sample(s.x1,y);
			}
			continue;
		}

		//section is a 'real' rectangle, at least three samples high and wide
		//get reference value from center of rectangle (for a somewhat nicer preview)
//		uint reference=fractal->get_result((s.x0+s.x1)>>1,(s.y0+s.y1)>>1);
		uint reference=(*fractal)(s.x0,s.y0);

		bool all_equal=true;
		//compare reference value to vertical edges
		for (x=s.x0;x<=s.x1;++x)
		{
			if (((*fractal)(x,s.y0)!=reference)||
				((*fractal)(x,s.y1)!=reference))
			{
				all_equal=false;
			}
		}
		for (y=s.y0;y<=s.y1;++y)
		{
			if (((*fractal)(s.x0,y)!=reference)||
				((*fractal)(s.x1,y)!=reference))
			{
				all_equal=false;
			}
		}
		if (all_equal)
		{
			filled_samples+=img->fill_interior(s,reference);
		}
		else
		{
			//split rectangle and add both halves to the queue
			sections.push_back(s.split());
			sections.push_back(s);
		}
	}
	return(ret);
}

float
BoxedGenerator::get_progress()
{
//FIXME
	float sampled=1.0f-fractal->get_cleverness();
	float filled=float(filled_samples)/(width*height);
	return(sampled+filled);
}

