/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * move HistEntry to seperate files - Rolf
03/Mar-05 * added load and save members to Context class - Rolf
03/Mar-05 * added proper LGPL notice, started changelog - Rolf
*/

class
Context
{
private:
	Context();	//hide default constructor
public:
	Context(Fractal* first_entry,uint megs);
	~Context();

	void reset_all();

	bool step_forward();
	bool step_back();
	Fractal* switch_view();

	const HistEntry* get_current() const {return(&sequence[current]);}
	bool refine_current();
	bool refine_others();
	const ubyte* get_indexed_image() const;
	const ubyte* get_rgb_image();

	int set_generator_type(int type)const;
	int get_generator_type()const;
	int change_generator_type()const;
	void set_image_properties(int x_res,int y_res,int aa_strength,float aa_gamma);
	bool toggle_preview();
	void reset_current_view();

	uint adjust_max_iterations(float factor);
	void zoom(int x,int y,float factor);
	void zoom(float factor);
	void move(float delta_x,float delta_y);
	
	void set_palette(const ubyte* const pal) { get_decoder()->set_palette(pal); }
	void set_palette_cycling_speed(float speed);

	bool load(const char* filename);
	bool save(const char* filename,bool everything);
	
private:
	RgbImage* get_decoder();
	void check_mem_usage();
	void duplicate_current();
	void make_generator(int entry);

	bool refine(int entry);
private:
	//image decoder
	RgbImage* decoder;
    RelTimer palette_cycle_timer;   //timer for palette cycling
    float palette_cycle_speed;      //speed of cycling, may be negative
    float cycle_time_remainder;     //rounding adjustment

	//controlled sequence of history entries
	HistEntry* sequence;
	int current;
	int newest;
	int range;			//size of the history sequence container

	uint total_memory;
	uint max_memory;

	//global setting for all history entries
	bool preview_enabled;
	float aa_gamma;
};

