/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

class
ExtremeFloat
{
public:
	ExtremeFloat();
	ExtremeFloat(const double a);
	ExtremeFloat(const int a);
	ExtremeFloat(const uint a);
	ExtremeFloat(const char* const c_string);

	void set_to_zero();

	//arithmetic
	void operator += (const ExtremeFloat& rhs);
	void operator -= (const ExtremeFloat& rhs);
	void operator *= (const ExtremeFloat& rhs);
	void operator *= (const uint rhs);
	void operator *= (const int rhs);

	//multiply by a power of 2
	void scale(const int x) { exponent+=x; }

	ExtremeFloat operator + (const ExtremeFloat& rhs) const;
	ExtremeFloat operator - (const ExtremeFloat& rhs) const;
	ExtremeFloat operator * (const ExtremeFloat& rhs) const;
	ExtremeFloat operator * (const int rhs) const;
	ExtremeFloat operator * (const uint rhs) const;

	const ExtremeFloat operator - () const {
		ExtremeFloat ret(*this);
		ret.sign^=1; return(ret); }

	//convert to IEEE-754 double precision
	double to_double() const;
	//convert to string
	void to_string(char* target_buffer) const;

	//comparison
	bool operator >= (const double& rhs) const;
	bool above_or_equal_pow2(const int exponent) const { return((0==sign)&&(this->exponent>=exponent)); }
	bool is_zero() const { return(0==mantissa[5]); }

private:
	uint mantissa[6];	//192 bits mantissa, explicit 'non-hidden bit'
	int exponent;		//zero based two's complement
	uint sign;			//single bit, 0=positive, 1=negative
};
