/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/




//stuff for understanding headers
#include <deque>			//needed for Generator.h
#include <vector>			//needed for MaskedImage.h
#include "common.h"
#include "ExtremeFloat.h"	//needed for Fractal.h
#include "Section.h"		//needed for MaskedImage.h
#include "Span.h"			//needed for MaskedImage.h
#include "MaskedImage.h"	//needed for Fractal.h
#include "Fractal.h"		//needed for Generator.h
#include "Generator.h"

//internal stuff
#include "x86_timer.ah"		//needed for Timing.h"
#include "Timing.h"

Generator::Generator(Fractal* fractal):
fractal(fractal),
preview_detail(8),
preview_enabled(true)
{
	width=fractal->get_width();
	height=fractal->get_height();
	reset_preview();
}

void
Generator::reset_preview()
{
	preview_data.clear();
	Section s(0,0,width-1,height-1);
	if ((s.x1>s.x0)&&(s.y1>s.y0))	preview_data.push_front(s);
	preview_finished=false;
}

void
Generator::notify_resize()
{
	if ((width!=fractal->get_width())||(height!=fractal->get_height()))
	{
		this->width=width;
		this->height=height;
		reset();
	}
}

void
Generator::set_preview_status(bool allowed)
{
    //restart, a newly started preview run may ruin some of our work
    if ((!preview_finished)&&(allowed)&&(!this->preview_enabled)) reset();
	preview_enabled=allowed;
}

//this is now a fullblown rectangular optimized kaboom generator device thingy (TM)
//see Boxed.cpp for more comments, it's almost the same
bool
Generator::preview(float max_time)
{
	if ((!preview_enabled)||preview_finished) return(false);

	const MaskedImage* img=fractal->get_image();

	AbsTimer preview_timer;

	while ((!preview_data.empty())&&(preview_timer.elapsed_seconds()<max_time))
	{
		Section s=preview_data.front();
		preview_data.pop_front();

		uint reference=(*fractal)(s.x0,s.y0);

		if (((s.y1-s.y0)<preview_detail)&&
			((s.x1-s.x0)<preview_detail))
		{
			img->preview_fill(s,reference);
			continue;
		}

		bool all_equal=true;

		for (uint x=s.x0;x<=s.x1;x+=preview_detail)
		{
			int r1=(*fractal)(x,s.y0);
			int r2=(*fractal)(x,s.y1);

			if ((r1!=reference)||
				(r2!=reference))
			{
				all_equal=false;
				break;
			}
		}
		for (uint y=s.y0;y<=s.y1;y+=preview_detail)
		{
			int r1=(*fractal)(s.x0,y);
			int r2=(*fractal)(s.x1,y);

			if ((r1!=reference)||
				(r2!=reference))
			{
				all_equal=false;
				break;
			}
		}
		if (all_equal)
		{
			img->preview_fill(s,reference);
		}
		else
		{
			img->preview_outline(s,reference);
			//split rectangle and add both halves to the queue
			Section other_half=s.split();
			preview_data.push_back(s);
			preview_data.push_back(other_half);
		}
	}
	if (preview_data.empty()) preview_finished=true;
	return(true);
}
