/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-06 * new file - Rolf
*/


class
TreeSymbol
{
public:
	TreeSymbol():value(0),code(0),code_len(0) {};

	void append_bit(uint bit){ code<<=1; code|=bit; ++code_len;  }

	ushort code;			//code for the encoded symbol
	ubyte code_len;			//length of the code in bits
	ubyte value;			//the actual encoded symbol
};

class
Tree
{
public:
	Tree();
	Tree(uint symbol,uint weight):sub0(NULL),sub1(NULL),weight(weight),data(symbol) {};
	Tree(Tree* child0,Tree* child1):sub0(child0),sub1(child1),data(1)
		{weight=sub0->weight+sub1->weight; sub1->notch_down(); sub0->notch_down();}

	~Tree() { if (sub0) { delete(sub0); delete(sub1); } }

	uint traverse(const uint code)const { if (sub0==NULL) return(data);	else
		{
			if (code&data)	return(sub1->traverse(code));
			else			return(sub0->traverse(code));
		}
	}

	void children_add_bit(uint bit,TreeSymbol* array) {
		if (NULL==sub0) array[data].append_bit(bit);
		else
		{
			sub0->children_add_bit(bit,array);
			sub1->children_add_bit(bit,array);
		}
	}

	Tree* sub0;			//NULL indicates a leaf
	Tree* sub1;			//should also be NULL _only_ for leaves, but this condition is not checked
	uint data;			//encoded symbol for leaves; and-mask for non-leaves
	uint weight;		//used for sorting. also ensures good alignment
private:
	//moves data one bit to the left
 	//in non-leaf nodes, data can be abused as an 'and'-mask to find the next traversal direction
	void notch_down(){ if (sub0) { data<<=1; sub0->notch_down(); sub1->notch_down(); } }
};

class
HuffmanCode
{
public:
	HuffmanCode();
	~HuffmanCode();

	const TreeSymbol* get_code(ubyte symbol)const {return(syms+symbol); }

//	const TreeSymbol* extract_symbol(uint code)const {return(syms+tree->traverse(code)); }
	
	const TreeSymbol* extract_symbol(uint code)const{ const Tree* t=tree;
		while (t->sub0) {if ((code&1)==0) t=t->sub0; else t=t->sub1; code>>=1; }
		return(syms+t->data); }

private:
	int max_symbols;
	Tree* tree;						//for finding the symbol to a code (decoding)
	TreeSymbol syms[256];			//for finding the code to a symbol (encoding)
};

