/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/


#include <deque>					//needed for Generator.h
#include <vector>					//needed for ContextFactory.h
#include "common.h"
#include "Section.h"				//needed for MaskedImage.h
#include "Span.h"					//needed for MaskedImage.h
#include "downfilter.ah"			//needed for RgbImage.h
#include "x86_timer.ah"				//needed for RgbImage.h
#include "Timing.h"					//needed for RgbImage.h
#include "MaskedImage.h"			//needed for Fractal.h
#include "ExtremeFloat.h"			//needed for Fractal.h
#include "Fractal.h"				//needed for Context.h
#include "Generator.h"				//needed for Context.h
#include "RgbImage.h"				//needed for Context.h

#include "brot_core_interface.h"	//needed for Context.h

#include "HistEntry.h"
#include "Context.h"
#include "ContextFactory.h"


ContextFactory cf;

#define CHECK_CONTEXT_void if (!cf.is_context(c)) { return; }
#define CHECK_CONTEXT(return_val_on_failure) if (!cf.is_context(c)) { return(return_val_on_failure); }

BROT_FUNCTION(bc_create_context,Context*,(int megs))
{
	if (megs<0) megs=32;
	return(cf.create_context(new Fractal(100,100),megs));
}

BROT_FUNCTION(bc_reset_context,void,(Context* c))
{
	CHECK_CONTEXT_void;
	c->reset_all();
}

BROT_FUNCTION(bc_delete_context,void,(Context** c))
{
	if (!cf.is_context(*c)) return;
	cf.destroy(*c);
	*c=0;
}

BROT_FUNCTION(bc_refine_image,bool,(Context* c))
{
	CHECK_CONTEXT(false);
	return(c->refine_current());
}

BROT_FUNCTION(bc_refine_history,bool,(Context* c))
{
	CHECK_CONTEXT(false);
	return(c->refine_others());
}

BROT_FUNCTION(bc_set_generator_type,int,(Context* c,int type))
{
	CHECK_CONTEXT(BC_CONTOUR_TRACING);
	return(c->set_generator_type(type));
}

BROT_FUNCTION(bc_change_generator_type,int,(Context* c))
{
	CHECK_CONTEXT(BC_CONTOUR_TRACING);
	return(c->change_generator_type());
}

BROT_FUNCTION(bc_set_image_properties,void,(Context* c,int width,int height,int aa_level,float aa_gamma))
{
	CHECK_CONTEXT_void;
	c->set_image_properties(width,height,aa_level,aa_gamma);
}

BROT_FUNCTION(bc_set_palette,void,(Context* c,const ubyte* const palette))
{
	CHECK_CONTEXT_void;
	c->set_palette(palette);
}

BROT_FUNCTION(bc_set_palette_cycling_speed,void,(Context* c,float speed))
{
    CHECK_CONTEXT_void;
    c->set_palette_cycling_speed(speed);
}

BROT_FUNCTION(bc_toggle_preview,bool,(Context* c))
{
	CHECK_CONTEXT(false);
	return(c->toggle_preview());
}

BROT_FUNCTION(bc_adjust_maximum_iterations,uint,(Context* c,float factor))
{
	CHECK_CONTEXT(0);
	return(c->adjust_max_iterations(factor));
}

BROT_FUNCTION(bc_get_maximum_iterations,uint,(Context* c))
{
	CHECK_CONTEXT(0);
	return(c->adjust_max_iterations(1.0f));
}

BROT_FUNCTION(bc_describe_generator,const char*const,(int type))
{
	switch(type)
	{
	case(BC_CONTOUR_TRACING):
		return("contour tracing");
		break;
	case(BC_BOX_MATCHING):
		return("box matching");
		break;
	case(BC_BRUTE_FORCE):
		return("brute force");
		break;
	default:
		return("unknown!?");
		break;
	}
}

BROT_FUNCTION(bc_zoom_on_window_coord,void,(Context* c,int x,int y,float factor))
{
	CHECK_CONTEXT_void;
	if ((factor<=0.0f)) factor=1.0f;
	c->zoom(x,y,factor);
}

BROT_FUNCTION(bc_zoom_centered,void,(Context* c,float factor))
{
	CHECK_CONTEXT_void;
	if (factor<=0.0f) return;
	c->zoom(factor);
}

BROT_FUNCTION(bc_move,void,(Context* c,float delta_x,float delta_y))
{
	CHECK_CONTEXT_void;
	c->move(delta_x,delta_y);
}

BROT_FUNCTION(bc_get_rgb_image,const ubyte*,(Context* c))
{
	CHECK_CONTEXT(NULL);
	return(c->get_rgb_image());
}

BROT_FUNCTION(bc_get_image_width,int,(Context* c))
{
	CHECK_CONTEXT(0);
	return(c->get_current()->width);
}

BROT_FUNCTION(bc_get_image_height,int,(Context* c))
{
	CHECK_CONTEXT(0);
	return(c->get_current()->height);
}

BROT_FUNCTION(bc_get_cleverness,float,(Context* c))
{
	CHECK_CONTEXT(0.0f);
	return(c->get_current()->cleverness);
}

BROT_FUNCTION(bc_get_progress,float,(Context* c))
{
	CHECK_CONTEXT(0.0f);
	return(c->get_current()->progress);
}

BROT_FUNCTION(bc_get_time_spent,float,(Context* c))
{
    CHECK_CONTEXT(0.0f);
    return(c->get_current()->total_time);
}

BROT_FUNCTION(bc_reset_current_view,void,(Context* c))
{
	CHECK_CONTEXT_void;
	c->reset_current_view();
}

BROT_FUNCTION(bc_history_back,bool,(Context* c))
{
	CHECK_CONTEXT(false);
	return(c->step_back());
}

BROT_FUNCTION(bc_history_forward,bool,(Context* c))
{
	CHECK_CONTEXT(false);
	return(c->step_forward());
}

struct Real_place
{
	ExtremeFloat real;
	ExtremeFloat img;
	ExtremeFloat radius;
	uint iterations;
};

BROT_FUNCTION(bc_store_location_to,void,(Context* c,void* place))
{
	Real_place* p=static_cast<Real_place*>(place);

	if (!cf.is_context(c))
	{
		p->real=0.0;
		p->img=0.0;
		p->radius=2.0;
		p->iterations=150;
	}
	else
	{
		p->real=c->get_current()->fractal->get_center_r();
		p->img=c->get_current()->fractal->get_center_i();
		p->radius=c->get_current()->fractal->get_radius();
		p->iterations=c->get_current()->fractal->get_max_iterations();
	}
}

BROT_FUNCTION(bc_restore_location_from,void,(Context* c,void* place))
{
	CHECK_CONTEXT_void;
	const Real_place* p=static_cast<Real_place*>(place);

	Fractal* f=c->switch_view();
	f->set_params(p->real,p->img,p->radius.to_double());
	f->set_max_iterations(p->iterations);
}

BROT_FUNCTION(bc_set_view_params_from_doubles,void,(Context* c,double real,double img,double radius))
{
	CHECK_CONTEXT_void;
	Fractal* f=c->switch_view();
	f->set_params(real,img,radius);
}

BROT_FUNCTION(bc_set_view_params_from_text,void,(Context* c,char* real,char* img,double radius))
{
	CHECK_CONTEXT_void;
	Fractal* f=c->switch_view();
	f->set_params(real,img,radius);
}

BROT_FUNCTION(bc_get_view_params_as_doubles,void,(Context* c,double* real,double* img,double* radius))
{
	if (!cf.is_context(c))
	{
		*real=0.0;
		*img=0.0;
		*radius=2.0;
	}
	else
	{
		*real=c->get_current()->fractal->get_center_r().to_double();
		*img=c->get_current()->fractal->get_center_i().to_double();
		*radius=c->get_current()->fractal->get_radius();
	}
}

BROT_FUNCTION(bc_get_view_params_as_text,void,(Context* c,char* real,char* img,double* radius))
{
	if (!cf.is_context(c))
	{
		strcpy(real,"0");
		strcpy(img,"0");
		*radius=2.0;
	}
	else
	{
		c->get_current()->fractal->get_center_r().to_string(real);
		c->get_current()->fractal->get_center_i().to_string(img);
		*radius=c->get_current()->fractal->get_radius();
	}
}

BROT_FUNCTION(bc_load,bool,(Context* c,const char* filename))
{
	CHECK_CONTEXT(false);
	return(c->load(filename));
}

BROT_FUNCTION(bc_save_all,bool,(Context* c,const char* filename))
{
	CHECK_CONTEXT(false);
	return(c->save(filename,true));
}

BROT_FUNCTION(bc_save_minimum,bool,(Context* c,const char* filename))
{
	CHECK_CONTEXT(false);
	return(c->save(filename,false));
}
