;this is part of brot_core.dll, a fractal generation library
;Copyright (C) 2003 Rolf Neuberger
;
;This library is free software; you can redistribute it and/or
;modify it under the terms of the GNU Lesser General Public
;License as published by the Free Software Foundation; either
;version 2.1 of the License, or (at your option) any later version.
;
;This library is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;Lesser General Public License for more details.
;
;You should have received a copy of the GNU Lesser General Public
;License along with this library; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


;changes:
;03/Mar-07 * now feeding the MMX result to ALU register through the stack - Rolf

SECTION .data

;we add this before we divide by 32 to achieve 'round to nearest'
rounding_thingy16: dw 16,16,16,16

SECTION .text

global _x86_downfilter_w_gamma_correction

;filter down a series of (indexed) pixels to an rgb result and apply
;gamma afterwards
;to reduce register pressure we use a struct with various lookup tables and data
;this struct is declared in downfilter.ah
;the palette must come with degamma preapplied and we assume that all subsample
;lookups added together fit a 16 bit integer

;prototype:
;void x86_downfilter_w_gamma_correction(ubyte* target_rgb,const ubyte* const src_indexed,
;										uint pixel_count,const downfilter_info* const di);

_x86_downfilter_w_gamma_correction:
	PUSHA
	MOV EDI,[ESP+36]
	MOV ESI,[ESP+40]
	MOV EBX,[ESP+44]
	MOV EBP,[ESP+48]

	SUB ESP,8							;allocate 8 bytes of temp storage on stack

	MOV ECX,[EBP+4352]
	LEA EDX,[EBP+4*ECX+4096]			;a handy pointer to the end of the offset array
.filter_loop:
	XOR ECX,ECX
	SUB ECX,[EBP+4352]					;get subpixel count (inverted loop counter)
	MOVQ mm0,[rounding_thingy16]		;clear r/g/b accumulators (preload rounding adjustment)
.accumulate_loop:
	MOVZX EAX,byte [ESI]				;get subpixel sample
	PADDW mm0,[EBP+8*EAX]				;accumulate samples (with degamma preapplied)
	ADD ESI,[EDX+4*ECX]					;adjust source pointer
	INC ECX
	JNZ .accumulate_loop

	PSRLW mm0,5							;divide by 32 with round to nearest
	MOVQ [ESP],mm0						;feed into ALU registers through stack
	MOVZX EAX,word [ESP]				;red
	MOVZX ECX,word [ESP+2]				;green

	MOV AL,[EBP+EAX+2048]				;perform gamma lookups and write out
	MOV AH,[EBP+ECX+2048]

	MOVZX ECX,word [ESP+4]				;blue

	MOV [EDI],AL
	MOV [EDI+1],AH
	MOV AL,[EBP+ECX+2048]
	MOV [EDI+2],AL

	ADD EDI,3
	DEC EBX
	JNZ .filter_loop

	ADD ESP,8							;deallocate temp storage

	EMMS
	POPA
	RETN
