;this is part of brot_core.dll, a fractal generation library
;Copyright (C) 2003 Rolf Neuberger
;
;This library is free software; you can redistribute it and/or
;modify it under the terms of the GNU Lesser General Public
;License as published by the Free Software Foundation; either
;version 2.1 of the License, or (at your option) any later version.
;
;This library is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;Lesser General Public License for more details.
;
;You should have received a copy of the GNU Lesser General Public
;License along with this library; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


;add second operand (which must have larger magnitude) to first operand
;prototype:
;void x86_xfloat_abs_add(ExtremeFloat* target,const ExtremeFloat* operand)
_x86_xfloat_abs_add:
	PUSHA
	MOV ESI,[ESP+36]
	MOV EDI,[ESP+40]
	MOV ECX,[EDI+24]		;get large exponent
	MOV EDX,[ESI+24]		;get small exponent
	MOV [ESI+24],ECX		;copy large exponent
	;EAX will store first bit shifted out (used for rounding)
	;if we don't shift at all, we need a zero in here
	XOR EAX,EAX
	PXOR mm5,mm5			;we use mm5 as a zero mask in several places
	SUB ECX,EDX				;get exponent delta
	;skip mantissa shifting for equal exponents
	JZ .add_mants
	CMP ECX,192
	JB .pre_shift
	JZ .round_up
	;we can't add anything for an exponent delta above 192
	;we'll just copy the larger operand
	MOVQ mm0,[EDI]
	MOVQ mm1,[EDI+8]
	MOVQ mm2,[EDI+16]
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm1
	MOVQ [ESI+16],mm2
	JMP .finished
.round_up:
	;exponent delta is exactly 192
	;we need to round up
	INC EAX					;carry=one
	;wipe original mantissa to zero
	PXOR mm0,mm0
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm0
	MOVQ [ESI+16],mm0
	JMP .add_mants

.pre_shift:
;split (shift count-1) into DWORDs and bits
	MOV EDX,ECX
	DEC ECX
	MOV EAX,ECX
	AND ECX,31
	SHR EAX,5
	MOV EAX,[ESI+4*EAX]
	SHR EAX,CL				;the 'lost' bit is now the lsb of EAX
;split shift count into qwords and bits
	MOV ECX,EDX
	AND ECX,63
	MOVD mm6,ECX
	NEG ECX
	ADD ECX,64
	MOVD mm7,ECX

	SHR EDX,6
	JNZ .shift_one_qword
	MOVQ mm0,[ESI]
	MOVQ mm1,[ESI+8]
	MOVQ mm2,[ESI+16]
	MOVQ mm3,mm1
	MOVQ mm4,mm2
	PSRLQ mm0,mm6
	PSRLQ mm1,mm6
	PSRLQ mm2,mm6
	PSLLQ mm3,mm7
	PSLLQ mm4,mm7
	POR mm0,mm3
	POR mm1,mm4
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm1
	MOVQ [ESI+16],mm2
	JMP .add_mants
.shift_one_qword:
	DEC EDX
	JNZ .shift_two_qwords
	MOVQ mm0,[ESI+8]
	MOVQ mm1,[ESI+16]
	MOVQ mm2,mm1
	PSRLQ mm0,mm6
	PSRLQ mm1,mm6
	PSLLQ mm2,mm7
	POR mm0,mm2
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm1
	MOVQ [ESI+16],mm5
	JMP .add_mants
.shift_two_qwords:
	MOVQ mm0,[ESI+16]
	PSRLQ mm0,mm6
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm5
	MOVQ [ESI+16],mm5

.add_mants:					;add mantissae
	SHR EAX,1				;get rounding adjustment into carry bit
	MOV EAX,[EDI]
	MOV EBX,[EDI+4]
	MOV ECX,[EDI+8]
	MOV EDX,[EDI+12]
	ADC [ESI],EAX
	ADC [ESI+4],EBX
	ADC [ESI+8],ECX
	ADC [ESI+12],EDX
	MOV EAX,[EDI+16]
	MOV EBX,[EDI+20]
	ADC [ESI+16],EAX
	ADC [ESI+20],EBX
;check_for_overflow
	JNC .finished
	;overflow, shift one place to the right
	RCR dword [ESI+20],1
	RCR dword [ESI+16],1
	RCR dword [ESI+12],1
	RCR dword [ESI+8],1
	RCR dword [ESI+4],1
	RCR dword [ESI],1

	INC dword [ESI+24]		;increase exponent
.finished:
	EMMS
	POPA
	RETN

;subtract first operand from second operand (which must have larger magnitude)
;and store result to first operand
;prototype:
;void x86_xfloat_abs_subr(ExtremeFloat* target,const ExtremeFloat* operand)
_x86_xfloat_abs_subr:
	PUSHA
	MOV ESI,[ESP+36]
	MOV EDI,[ESP+40]
	;EAX will store first two bits shifted out (used for rounding)
	;if we don't shift at all, we need a zero in here
	XOR EAX,EAX
	PXOR mm5,mm5				;we use mm5 as a zero mask in several places

	MOV ECX,[EDI+24]			;get large exponent
	MOV EDX,[ESI+24]			;get small exponent
	MOV [ESI+24],ECX			;copy large exponent
	SUB ECX,EDX					;subtract small exponent

	JZ .equal_exponents
	;exponents are not equal, we need to shift *this' mantissa to the right
	CMP ECX,1
	JZ .non_trivial
	CMP ECX,192
	JB .trivial_subtract

	;exponent delta is above or equal 192. We'll just copy the larger operand
	MOVQ mm0,[EDI]
	MOVQ mm1,[EDI+8]
	MOVQ mm2,[EDI+16]
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm1
	MOVQ [ESI+16],mm2
	JMP .finished

.trivial_subtract:
;exponents differ by more than one, but 191 at most
;this is good, because the result's exponent will at most change by one
	;split shift count into QWORDs/bits
	MOV EDX,ECX
	AND ECX,63
	MOVD mm6,ECX
	NEG ECX
	ADD ECX,64
	MOVD mm7,ECX

	SHR EDX,6
	MOVQ mm0,[ESI+8]		;load high 128 mantissa mantissa bits into mm1|mm0
	MOVQ mm1,[ESI+16]
	CMP EDX,1
	JA .shift_two_qwords
	JZ .shift_one_qword
;shift count is less than 64
	MOVQ mm2,[ESI]
	MOVQ mm3,mm0
	MOVQ mm4,mm1
	PSRLQ mm2,mm6
	PSRLQ mm3,mm6
	PSRLQ mm4,mm6
	PSLLQ mm0,mm7
	PSLLQ mm1,mm7
	POR mm2,mm0
	POR mm3,mm1
	MOVQ [ESI],mm2
	MOVQ [ESI+8],mm3
	MOVQ [ESI+16],mm4
	JMP .preshifted

.shift_one_qword:
	MOVQ mm2,mm1
	PSLLQ mm2,mm7
	PSRLQ mm0,mm6
	PSRLQ mm1,mm6
	POR mm0,mm2
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm1
	MOVQ [ESI+16],mm5
	JMP .preshifted

.shift_two_qwords:
	PSRLQ mm1,mm6
	MOVQ [ESI],mm1
	MOVQ [ESI+8],mm5
	MOVQ [ESI+16],mm5

.preshifted:
	MOV EAX,[EDI]
	MOV EBX,[EDI+4]
	MOV ECX,[EDI+8]
	SUB EAX,[ESI]
	SBB EBX,[ESI+4]
	SBB ECX,[ESI+8]
	MOV [ESI],EAX
	MOV [ESI+4],EBX
	MOV [ESI+8],ECX
	MOV EAX,[EDI+12]
	MOV EBX,[EDI+16]
	MOV ECX,[EDI+20]
	SBB EAX,[ESI+12]
	SBB EBX,[ESI+16]
	SBB ECX,[ESI+20]
	MOV [ESI+12],EAX
	MOV [ESI+16],EBX
	MOV [ESI+20],ECX
	JS .finished
;MSB is clear, need to shift left by one and decrement exponent
	SHL dword [ESI],1
	RCL dword [ESI+4],1
	RCL dword [ESI+8],1
	RCL dword [ESI+12],1
	RCL dword [ESI+16],1
	RCL dword [ESI+20],1

	DEC dword [ESI+24]		;adjust exponent
	JMP .finished

.non_trivial:
	;exponent delta is exactly one.
	;this produces unpredictable result exponents. We need to scan for msb after subtraction.
	XOR EAX,EAX
	MOV EBX,EAX
	MOV ECX,EAX
	MOV EDX,EAX
	NEG dword [ESI]
	SBB EBX,[ESI+4]
	SBB ECX,[ESI+8]
	SBB EDX,[ESI+12]
	MOV [ESI+4],EBX
	MOV EBX,EAX
	SBB EAX,[ESI+16]
	SBB EBX,[ESI+20]
	STC
	RCR EBX,1
	MOV [ESI+20],EBX
	RCR EAX,1
	MOV [ESI+16],EAX
	RCR EDX,1
	MOV [ESI+12],EDX
	RCR ECX,1
	MOV [ESI+8],ECX
		
	RCR dword [ESI+4],1
	RCR dword [ESI],1	;the carry flag now holds the lost bit (we always lose exactly one!)
	SETC AL
	JMP .subtract_nontrivial

.equal_exponents:
	XOR EAX,EAX
	MOV EBX,EAX
	MOV ECX,EAX
	MOV EDX,EAX
	NEG dword [ESI]
	SBB EBX,[ESI+4]
	SBB ECX,[ESI+8]
	SBB EDX,[ESI+12]
	MOV [ESI+4],EBX
	MOV [ESI+8],ECX
	MOV [ESI+12],EDX
	MOV EBX,EAX
	SBB EAX,[ESI+16]
	SBB EBX,[ESI+20]
	MOV [ESI+16],EAX
	MOV [ESI+20],EBX
;exponents are equal, we don't need to shift or round anything
	XOR EAX,EAX
.subtract_nontrivial:
	MOV EBX,[EDI]
	MOV ECX,[EDI+4]
	MOV EDX,[EDI+8]
	ADD [ESI],EBX
	ADC [ESI+4],ECX
	ADC [ESI+8],EDX
	MOV EBX,[EDI+12]
	MOV ECX,[EDI+16]
	MOV EDX,[EDI+20]
	ADC [ESI+12],EBX
	ADC [ESI+16],ECX
	ADC [ESI+20],EDX
	JS .finished				;if the msb is still set, we're finished
;we'll first try with shifting in on more bit, saved in AL
	DEC dword [ESI+24]		;mangle exponent
	SHR EAX,1					;get carry into carry flag (hrhr again)
	MOV EBX,[ESI]
	MOV ECX,[ESI+4]
	MOV EDX,[ESI+8]
	ADC [ESI],EBX
	ADC [ESI+4],ECX
	ADC [ESI+8],EDX
	MOV EBX,[ESI+12]
	MOV ECX,[ESI+16]
	MOV EDX,[ESI+20]
	ADC [ESI+12],EBX
	ADC [ESI+16],ECX
	ADC [ESI+20],EDX
	JS .finished
;find MSB of result
	;this might disgracefully crash if we wouldn't guarantee that the
	;absolute values of the two operands differ and are both nonzero
	MOV EDX,6
.find_msb_loop:
	DEC EDX
	BSR ECX,[ESI+4*EDX]
	JZ .find_msb_loop

	LEA EAX,[8*EDX]				;adjust exponent accordingly
	LEA EAX,[ECX+4*EAX-191]
	ADD [ESI+24],EAX

	NEG EAX
	MOV EDX,EAX
	AND EAX,63
	MOVD mm6,EAX
	NEG EAX
	ADD EAX,64
	MOVD mm7,EAX

	SHR EDX,6
	JNZ .post_shift_one_qword

	MOVQ mm0,[ESI]
	MOVQ mm1,[ESI+8]
	MOVQ mm2,[ESI+16]
	MOVQ mm3,mm0
	MOVQ mm4,mm1
	PSLLQ mm0,mm6
	PSLLQ mm1,mm6
	PSLLQ mm2,mm6
	PSRLQ mm3,mm7
	PSRLQ mm4,mm7
	POR mm1,mm3
	POR mm2,mm4
	MOVQ [ESI],mm0
	MOVQ [ESI+8],mm1
	MOVQ [ESI+16],mm2
	JMP .finished
.post_shift_one_qword:
	DEC EDX
	JNZ .post_shift_two_qwords
	MOVQ mm0,[ESI]
	MOVQ mm1,[ESI+8]
	MOVQ mm2,mm0
	PSLLQ mm0,mm6
	PSLLQ mm1,mm6
	PSRLQ mm2,mm7
	POR mm1,mm2
	MOVQ [ESI],mm5
	MOVQ [ESI+8],mm0
	MOVQ [ESI+16],mm1
	JMP .finished

.post_shift_two_qwords:
	MOVQ mm0,[ESI]
	PSLLQ mm0,mm6
	MOVQ [ESI],mm5
	MOVQ [ESI+8],mm5
	MOVQ [ESI+16],mm0

.finished:
	EMMS
	POPA
	RETN


